package ca.yorku.cse.mack.tiltmaze;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.res.Configuration;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.media.MediaScannerConnection;
import android.os.Bundle;
import android.os.CountDownTimer;
import android.os.Environment;
import android.os.Vibrator;
import android.util.DisplayMetrics;
import android.util.Log;
import android.view.Surface;
import android.view.WindowManager;

/**
 * TiltMaze
 *  
 * This is the game software used in...
 * 
 * @author Scott MacKenzie, Catalina Constantin, 2014
 */

public class TiltMazeActivity extends Activity implements SensorEventListener
{
	final static String MYDEBUG = "MYDEBUG"; // for Log.i messages

	final String WORKING_DIRECTORY = "/TiltMazeData/";
	final String APP = "TiltMaze";
	final String THE_HEADER = "App,Participant,Group,Session,Block,Order_Of_Control,Gain,Trial,"
			+ "Time_Total,Time_Off_Path,Wall_Hits,Movement_Nominal,Movement_Actual,Movement_Off_Path,Movement_Efficiency\n";
	final static int REFRESH_INTERVAL = 20; // milliseconds (screen refreshes @ 50 Hz)
	final float RADIANS_TO_DEGREES = 57.2957795f;
	final int NAVIGATION_BAR_HEIGHT = 48;

	// setup parameters
	String participant, session, group, block, orderOfControl;
	int gain;
	String gainLabel;
	int numberOfTrials;
	boolean showNominalPath, showMinimumPath, showWallHits, soundAlarmOnWallHit;

	// int constants to setup a sensor mode (see DemoTiltMeter API for discussion)
	final static int ORIENTATION = 0;
	final static int ACCELEROMETER_ONLY = 1;
	final static int ACCELEROMETER_AND_MAGNETIC_FIELD = 2;

	RollingBallPanel rb;

	/*
	 * Below are the alpha values for the low-pass filter. The four values are for the slowest
	 * (NORMAL) to fastest (FASTEST) sampling rates, respectively. These values were determined by
	 * trial and error. There is a trade-off. Generally, lower values produce smooth but sluggish
	 * responses, while higher values produced jerky but fast responses.
	 * 
	 * Furthermore, there is a difference by device, particularly for the FASTEST setting. For
	 * example, the FASTEST sample rate is about 200 Hz on a Nexus 4 but only about 100 Hz on a
	 * Samsung Galaxy Tab 10.1.
	 * 
	 * Fiddle with these, as necessary.
	 */
	final float[] ALPHA_VELOCITY = { 0.99f, 0.80f, 0.40f, 0.15f };
	final float[] ALPHA_POSITION = { 0.50f, 0.30f, 0.15f, 0.10f };
	float alpha;

	private SensorManager sm;
	private Sensor sA, sM, sO;
	int sensorMode;
	int defaultOrientation;

	float[] orientation = new float[3];
	float[] accValues = new float[3];
	float[] magValues = new float[3];
	float x, y, z, pitch, roll;

	ScreenRefreshTimer refreshScreen;

	// to get results && save data
	boolean showReportWindow;
	String base; // base filename for output data files
	String leader; // leader to identify conditions for data written to sd2 files.
	BufferedWriter writer;
	File theFile;

	// Called when the activity is first created
	@SuppressWarnings("deprecation")
	@Override
	public void onCreate(Bundle savedInstanceState)
	{
		super.onCreate(savedInstanceState);
		setContentView(R.layout.main);

		Bundle b = getIntent().getExtras();
		participant = b.getString(SetupActivity.PARTICIPANT_CODE_KEY);
		session = b.getString(SetupActivity.SESSION_CODE_KEY);
		group = b.getString(SetupActivity.GROUP_CODE_KEY);
		numberOfTrials = b.getInt(SetupActivity.TRIALS_CODE_KEY);
		orderOfControl = b.getString(SetupActivity.ORDER_OF_CONTROL_KEY);
		gain = b.getInt(SetupActivity.GAIN_KEY);
		gainLabel = b.getString(SetupActivity.GAIN_LABEL_KEY);
		showNominalPath = b.getBoolean(SetupActivity.SHOW_NOMINAL_PATH_KEY);
		showMinimumPath = b.getBoolean(SetupActivity.SHOW_MINIMUM_PATH_KEY);
		showWallHits = b.getBoolean(SetupActivity.SHOW_WALL_HITS_KEY);
		soundAlarmOnWallHit = b.getBoolean(SetupActivity.SOUND_ALARM_ON_WALL_HIT_KEY);

		// set alpha for low-pass filter (based on sampling rate and order of control)
		if (orderOfControl.equals("Velocity")) // velocity control
			alpha = ALPHA_VELOCITY[2]; // for GAME sampling rate
		else
			// position control
			alpha = ALPHA_POSITION[2]; // for GAME sampling rate

		// determine screen width, screen height, and pixel density
		DisplayMetrics dm = new DisplayMetrics();
		getWindowManager().getDefaultDisplay().getMetrics(dm);
		int w = dm.widthPixels;
		int h = dm.heightPixels;

		defaultOrientation = getDefaultDeviceOrientation();
		if (defaultOrientation == Configuration.ORIENTATION_LANDSCAPE)
			h -= NAVIGATION_BAR_HEIGHT;

		float pixelDensity = dm.density;

		showReportWindow = false;

		// initialize the output data file (and write the header line)
		fileInit();

		// configure rolling ball panel, as per screen parameters and setup parameters
		rb = (RollingBallPanel)findViewById(R.id.rollingballpanel);
		rb.configure((float)w, (float)h, pixelDensity);
		rb.setNumberOfTrials(numberOfTrials);
		rb.setGain(gain);
		rb.setOrderOfControl(orderOfControl);
		rb.setShowNominalPath(showNominalPath);
		rb.setShowMinimumPath(showMinimumPath);
		rb.setShowWallHits(showWallHits);
		rb.setSoundAlarmOnWallHit(soundAlarmOnWallHit);
		rb.setBlockCode(block);
		rb.setVibrator((Vibrator)getSystemService(Context.VIBRATOR_SERVICE));

		// get sensors
		sm = (SensorManager)getSystemService(SENSOR_SERVICE);
		sO = sm.getDefaultSensor(Sensor.TYPE_ORIENTATION); // supported on many devices
		sA = sm.getDefaultSensor(Sensor.TYPE_ACCELEROMETER); // supported on most devices
		sM = sm.getDefaultSensor(Sensor.TYPE_MAGNETIC_FIELD); // null on many devices

		// setup the sensor mode (see DemoTiltMeterAPI for discussion)
		if (sO != null)
		{
			sensorMode = ORIENTATION;
			sA = null;
			sM = null;
			Log.i(MYDEBUG, "Sensor mode: ORIENTATION");
		} else if (sA != null && sM != null)
		{
			sensorMode = ACCELEROMETER_AND_MAGNETIC_FIELD;
			Log.i(MYDEBUG, "Sensor mode: ACCELEROMETER_AND_MAGNETIC_FIELD");
		} else if (sA != null)
		{
			sensorMode = ACCELEROMETER_ONLY;
			Log.i(MYDEBUG, "Sensor mode: ACCELEROMETER_ONLY");
		} else
		{
			Log.i(MYDEBUG, "Can't run demo.  Requires Orientation sensor or Accelerometer");
			this.finish();
		}

		// NOTE: sensor listeners are registered in onResume

		// setup the screen refresh timer (screen updates every REFRESH_INTERVAL milliseconds)
		refreshScreen = new ScreenRefreshTimer(REFRESH_INTERVAL, REFRESH_INTERVAL);
		refreshScreen.start();
	}

	/** Creates the file for data recording */
	private void fileInit()
	{
		// ===================
		// File initialization
		// ===================

		// make a working directory (if necessary) to store data files
		File dataDirectory = new File(Environment.getExternalStorageDirectory() + WORKING_DIRECTORY);
		if (!dataDirectory.exists() && !dataDirectory.mkdirs())
		{
			Log.e(MYDEBUG, "ERROR --> FAILED TO CREATE DIRECTORY: " + WORKING_DIRECTORY);
			super.onDestroy(); // cleanup
			this.finish(); // terminate
		}

		// base filename for output data files (find first available block code and use in filename)
		int blockNumber = 1;
		block = "B01";
		base = APP + "-" + participant + "-" + session + "-" + block + "-" + group + "-" + orderOfControl;

		theFile = new File(dataDirectory, base + ".csv");

		// if file exists, increment the block number and try again
		while (theFile.exists())
		{
			++blockNumber;
			block = blockNumber < 10 ? "B0" + blockNumber : "B" + blockNumber;
			base = APP + "-" + participant + "-" + session + "-" + block + "-" + group + "-" + orderOfControl;
			theFile = new File(dataDirectory, base + ".csv");
		}

		/*
		 * Make a comma-delimited leader (for importing into Excel). The leader is pre-appended to
		 * the data written to the data file (see below).
		 */
		leader = APP + "," + participant + "," + group + "," + session + "," + block + "," + orderOfControl + ","
				+ gainLabel;

		try
		{
			writer = new BufferedWriter(new FileWriter(theFile));

			// output header in sd1 file
			writer.write(THE_HEADER, 0, THE_HEADER.length());
			writer.flush();
		} catch (IOException e)
		{
			Log.e(MYDEBUG, "ERROR OPENING DATA FILE! e=" + e.toString());
			super.onDestroy();
			this.finish();
		} // end file initialization
	}

	/*
	 * Get the default orientation of the device. This is needed to correctly map the sensor data
	 * for pitch and roll (see onSensorChanged).
	 */
	public int getDefaultDeviceOrientation()
	{
		WindowManager windowManager = (WindowManager)getSystemService(WINDOW_SERVICE);
		Configuration config = getResources().getConfiguration();
		int rotation = windowManager.getDefaultDisplay().getRotation();

		if (((rotation == Surface.ROTATION_0 || rotation == Surface.ROTATION_180) && config.orientation == Configuration.ORIENTATION_LANDSCAPE)
				|| ((rotation == Surface.ROTATION_90 || rotation == Surface.ROTATION_270) && config.orientation == Configuration.ORIENTATION_PORTRAIT))
			return Configuration.ORIENTATION_LANDSCAPE;
		else
			return Configuration.ORIENTATION_PORTRAIT;
	}

	@Override
	protected void onResume()
	{
		super.onResume();
		sm.registerListener(this, sO, SensorManager.SENSOR_DELAY_GAME); // good enough!
		sm.registerListener(this, sA, SensorManager.SENSOR_DELAY_GAME);
		sm.registerListener(this, sM, SensorManager.SENSOR_DELAY_GAME);
	}

	@Override
	protected void onPause()
	{
		super.onPause();
		sm.unregisterListener(this);
	}

	@Override
	public void onAccuracyChanged(Sensor sensor, int accuracy)
	{
		// not needed but we need to provide an implementation anyway
	}

	@Override
	public void onSensorChanged(SensorEvent se)
	{
		// ===============================
		// DETERMINE DEVICE PITCH AND ROLL
		// ===============================

		switch (sensorMode)
		{
			case ORIENTATION: // ========================================================

				/*
				 * Use this mode if the device has an orientation sensor.
				 */

				if (se.sensor.getType() != Sensor.TYPE_ORIENTATION)
				{
					Log.i(MYDEBUG, "Return now. Sensor event from " + se.sensor.getName());
					return;
				}

				// This bit of fiddling is necessary so the app will work on different devices.
				switch (defaultOrientation)
				{
					case Configuration.ORIENTATION_PORTRAIT:
					{
						// e.g., Nexus 4
						pitch = se.values[1];
						roll = se.values[2];
						break;
					}
					case Configuration.ORIENTATION_LANDSCAPE:
					{
						// e.g., Samsung Galaxy Tab 10.1
						pitch = se.values[2];
						roll = -se.values[1];
						break;
					}
				}
				break;

			case ACCELEROMETER_AND_MAGNETIC_FIELD: // ===================================

				/*
				 * Use this mode if the device has both an accelerometer and a magnetic field sensor
				 * (but no orientation sensor). See...
				 * 
				 * http://blog.thomnichols.org/2012/06/smoothing-sensor-data-part-2
				 */

				// smooth the sensor values using a low-pass filter
				if (se.sensor.getType() == Sensor.TYPE_ACCELEROMETER)
					accValues = lowPass(se.values.clone(), accValues, alpha);
				if (se.sensor.getType() == Sensor.TYPE_MAGNETIC_FIELD)
					magValues = lowPass(se.values.clone(), magValues, alpha);

				if (accValues != null && magValues != null)
				{
					// compute pitch and roll
					float R[] = new float[9];
					float I[] = new float[9];
					boolean success = SensorManager.getRotationMatrix(R, I, accValues, magValues);
					if (success) // see SensorManager API
					{
						float[] orientation = new float[3];
						SensorManager.getOrientation(R, orientation); // see getOrientation API
						pitch = orientation[1] * RADIANS_TO_DEGREES;
						roll = -orientation[2] * RADIANS_TO_DEGREES;
					}
				}
				break;

			case ACCELEROMETER_ONLY: // =================================================

				/*
				 * Use this mode if the device has an accelerometer but no magnetic field sensor and
				 * no orientation sensor (e.g., HTC Desire C, Asus MeMOPad). This algorithm doesn't
				 * work quite as well, unfortunately. See...
				 * 
				 * http://www.hobbytronics.co.uk/accelerometer-info
				 */

				// smooth the sensor values using a low-pass filter
				if (se.sensor.getType() == Sensor.TYPE_ACCELEROMETER)
					accValues = lowPass(se.values.clone(), accValues, alpha);

				x = accValues[0];
				y = accValues[1];
				z = accValues[2];
				pitch = (float)Math.atan(y / Math.sqrt(x * x + z * z)) * RADIANS_TO_DEGREES;
				roll = (float)Math.atan(x / Math.sqrt(y * y + z * z)) * RADIANS_TO_DEGREES;
				break;
		}
	}

	/*
	 * Low pass filter. The algorithm requires tracking only two numbers - the prior number and the
	 * new number. There is a time constant "alpha" which determines the amount of smoothing. Alpha
	 * is like a "weight" or "momentum". It determines the effect of the new value on the current
	 * smoothed value.
	 * 
	 * A lower alpha means more smoothing. NOTE: 0 <= alpha <= 1.
	 * 
	 * See...
	 * 
	 * http://blog.thomnichols.org/2011/08/smoothing-sensor-data-with-a-low-pass-filter
	 */
	protected float[] lowPass(float[] input, float[] output, float alpha)
	{
		for (int i = 0; i < input.length; i++)
			output[i] = output[i] + alpha * (input[i] - output[i]);
		return output;
	}

	/*
	 * Screen updates are done in onFinish which executes every REFRESH_INTERVAL milliseconds
	 */
	public class ScreenRefreshTimer extends CountDownTimer
	{
		ScreenRefreshTimer(long millisInFuture, long countDownInterval)
		{
			super(millisInFuture, countDownInterval);
		}

		@Override
		public void onTick(long millisUntilFinished)
		{
		}

		@Override
		public void onFinish()
		{
			// get the device's tilt magnitude and angle (from the sensor's pitch and roll)
			float tiltMagnitude = (float)Math.sqrt(pitch * pitch + roll * roll);
			float tiltAngle = tiltMagnitude == 0f ? 0f : (float)Math.asin(roll / tiltMagnitude) * RADIANS_TO_DEGREES;

			// adjust if necessary
			if (pitch > 0 && roll > 0)
				tiltAngle = 360f - tiltAngle;
			else if (pitch > 0 && roll < 0)
				tiltAngle = -tiltAngle;
			else if (pitch < 0 && roll > 0)
				tiltAngle = tiltAngle + 180f;
			else if (pitch < 0 && roll < 0)
				tiltAngle = tiltAngle + 180f;

			// check the ball panel to see if we are done
			if (rb.done && !showReportWindow)
			{
				// Done! show/store results
				showReport();
				showReportWindow = true;
				recordResults();
			} else
			{
				/*
				 * Not done! update the ball position, etc. This will invalidate the ball panel to
				 * force a redraw. Restart the timer to prepare for next update.
				 */
				rb.updateBallPosition(pitch, roll, tiltAngle, tiltMagnitude);
				this.start();
			}
		}
	}

	// Shows the Report window and allows for exit or restart of the application
	private void showReport()
	{
		float totalTimeForTrial = 0f;
		float totalTimeOffPath = 0f;
		int totalWallHits = 0;
		float totalMovementNominal = 0f;
		float totalMovementActual = 0f;

		for (int i = 0; i < rb.data.length; ++i)
		{
			totalTimeForTrial += rb.data[i].timeForTrialResult;
			totalTimeOffPath += rb.data[i].timeOffPathResult;
			totalWallHits += rb.data[i].wallHitsResult;
			totalMovementNominal += rb.data[i].movementNominalResult;
			totalMovementActual += rb.data[i].movementActualResult;
		}

		// bundle up parameters to pass on to results activity
		Bundle b = new Bundle();
		b.putString("block", this.block);
		b.putInt("numTrials", rb.numberOfTrials);
		b.putFloat("time", totalTimeForTrial);
		b.putFloat("timeOffPath", totalTimeOffPath);
		b.putInt("wallHits", totalWallHits);
		b.putFloat("movementEfficiency", totalMovementNominal / totalMovementActual * 100.0f);

		// start results activity
		Intent i = new Intent(getApplicationContext(), ResultsActivity.class);
		i.putExtras(b);
		startActivity(i);
		finish();
	}

	// Record the user results
	private void recordResults()
	{
		StringBuilder theData = new StringBuilder(1000);

		for (int i = 0; i < rb.data.length; ++i)
		{
			theData.append(leader + ",");
			theData.append(rb.data[i].trialNumberResult + ",");
			theData.append(rb.data[i].timeForTrialResult + ",");
			theData.append(rb.data[i].timeOffPathResult + ",");
			theData.append(rb.data[i].wallHitsResult + ",");
			theData.append(rb.data[i].movementNominalResult + ",");
			theData.append(rb.data[i].movementActualResult + ",");
			theData.append(rb.data[i].movementOffPathResult + ",");
			theData.append(rb.data[i].movementEfficiencyResult + "\n");
		}
		try
		{
			writer.write(theData.toString(), 0, theData.length());
		} catch (IOException e)
		{
			Log.e(MYDEBUG, "ERROR WRITING TO DATA FILE! e=" + e.toString());
		}

		doDone();
	}

	// Closes the file and makes sure it's visible
	private void doDone()
	{
		try
		{
			writer.close();
			MediaScannerConnection.scanFile(this, new String[] { theFile.getAbsolutePath() }, null, null);

		} catch (IOException e)
		{
			Log.e(MYDEBUG, "ERROR CLOSING DATA FILE! e=" + e);
		}
	}
}